# -*- coding: utf-8 -*-
#
# This file is part of s4d.
#
# s4d is a python package for speaker diarization.
# Home page: http://www-lium.univ-lemans.fr/s4d/
#
# s4d is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# s4d is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with s4d.  If not, see <http://www.gnu.org/licenses/>.


"""
Copyright 2014-2020 Sylvain Meignier
"""

# import matplotlib.pyplot as plt
# import time
# import os
#
# from ecyglpki import Problem, SimplexControls
# from scipy.cluster import hierarchy as hac
# from .hac_utils import *


# class ILP_IV:
#     """
#
#     """
#     def __init__(self, diar, scores, threshold=0.0):
#         """
#
#         :param diar:
#         :param scores:
#         :param threshold:
#         """
#         self.sep = '!!'
#         self.dic = dict()
#         self.diar = copy.deepcopy(diar)
#
#         self.thr = threshold
#         self.scores = copy.deepcopy(scores)
#
#     def plot(self, link, cluster_list):
#         """
#
#         :param link:
#         :param cluster_list:
#         :return:
#         """
#         #print(link)
#         hac.dendrogram(link, color_threshold=self.thr, labels=cluster_list)
#         plt.show()
#
#     def perform(self, save_ilp=False):
#         """
#         Creates an LP problem and solves it using GLPK library
#         (with ecyglpki Cython interface)
#         :param save_ilp: if True, saves the generated problem
#         in CPLEX LP format (problem.ilp), and the solution of
#         the problem (solution.ilp).
#         :return: a Diar object with the new clusters and a
#         dictionary of clusters (old clusters as key, new ones
#         as value in an array).
#         """
#         distances, t = scores2distance(self.scores, self.thr)
#         cluster_list = sorted(self.scores.modelset.tolist())
#
#         lp = Problem()
#         lp.set_obj_dir('minimize')
#
#         # columns
#         cols = []
#         for idx, cluster in enumerate(cluster_list):
#             col = '{}{}{}'.format(cluster, self.sep, cluster)
#             cols.append(col)
#             lp.add_named_cols(col)
#             lp.set_obj_coef(col, 1)
#             lp.set_col_bnds(col, 0, None)
#
#         # sum of dist > thr in the lower triangular part od the distance matrix
#         mask = (np.tril(distances, -1) > t)
#         threshold = np.tril(distances, -1).copy()
#         threshold[mask] = 0
#         s = np.sum(threshold) + 1
#         l = len(cluster_list)
#         for i in range(l):
#             cluster_i = cluster_list[i]
#             for j in range(i + 1, l):
#                 if distances[i, j] < t:
#                     cluster_j = cluster_list[j]
#                     v = distances[i, j] / s
#                     col = '{}{}{}'.format(cluster_i, self.sep, cluster_j)
#                     cols.append(col)
#                     lp.add_named_cols(col)
#                     lp.set_obj_coef(col, v)
#                     lp.set_col_bnds(col, 0, None)
#
#         # rows
#         for i, cluster_i in enumerate(cluster_list):
#             r_cols = {}
#             row = 'S{}'.format(i)
#             lp.add_named_rows(row)
#             r_cols['{}{}{}'.format(cluster_i, self.sep, cluster_i)] = 1
#
#             for j, cluster_j in enumerate(cluster_list):
#                 if i != j and distances[i, j] < t:
#                     col = '{}{}{}'.format(cluster_i, self.sep, cluster_j)
#                     if col not in cols:
#                         cols.append(col)
#                         lp.add_named_cols(col)
#                         lp.set_col_bnds(col, 0, None)
#                     r_cols[col] = 1
#
#                     #boundaries <= 0
#                     col = '{}{}{}'.format(cluster_i, self.sep, cluster_j)
#                     idx = lp.add_rows(1)
#                     lp.set_mat_row(idx, {'{}{}{}'.format(cluster_j, self.sep, cluster_j):-1, col:1})
#                     lp.set_row_bnds(idx, None, 0)
#
#             lp.set_mat_row(row, r_cols)
#             lp.set_row_bnds(row, 1, 1)
#
#         if save_ilp:
#             lp.write_lp('problem.ilp')
#
#         # solving problem
#         ctrl = SimplexControls()
#         ctrl.presolve = True
#         lp.simplex(ctrl)
#
#         if save_ilp:
#             lp.print_sol('solution.ilp')
#
#         cluster_dict = dict()
#         for i in range(lp.get_num_cols()):
#             names = lp.get_col_name(i+1).split(self.sep)
#             activity = lp.get_col_prim(i+1)
#             if activity == 1 and names[0] != names[1]:
#                 if names[1] not in cluster_dict:
#                     cluster_dict[names[1]] = []
#                 cluster_dict[names[1]].append(names[0])
#
#         table = copy.deepcopy(self.diar)
#         for idx in cluster_dict:
#             table.rename('cluster', cluster_dict[idx], idx)
#         return table, cluster_dict
#
#
#
#     def _perform(self, filename='tmp.ilp', rm_tmp=True):
#         """
#         Same as perform(), using glpk solver directly
#         instead of the Cython interface (slower).
#         """
#         table = copy.deepcopy(self.diar)
#         logging.debug('ilp filename: %s', filename)
#         f = open(filename, 'w')
#         self._ilp_write(f)
#         f.close()
#         while not os.path.exists(filename):
#             time.sleep(1)
#         cmd = 'glpsol --lp {} -o {} &> {}'.format(filename, filename + '.out', filename + '.err')
#         # print(cmd)
#         if os.path.exists(filename + '.out'):
#             os.remove(filename + '.out')
#             while os.path.exists(filename + '.out'):
#                 time.sleep(1)
#         os.system(cmd)
#         time.sleep(1)
#         while not os.path.exists(filename + '.out'):
#             time.sleep(1)
#         f = open(filename + '.out', 'r')
#         cluster_dict = self._ilp_read(f)
#         f.close()
#         for idx in cluster_dict:
#             table.rename('cluster', cluster_dict[idx], idx)
#         if rm_tmp:
#             os.remove(filename)
#             os.remove(filename+ '.out')
#             os.remove(filename+ '.err')
#             while os.path.exists(filename) or os.path.exists(filename+ '.out') or os.path.exists(filename+ '.err'):
#                 time.sleep(1)
#         return table, cluster_dict
#
#     def _ilp_write(self, f):
#         distances, t = scores2distance(self.scores, self.thr)
#
#         cluster_list = self.scores.modelset.tolist()
#         f.write("Minimize\n")
#         f.write("problem : \n")
#         for idx, cluster in enumerate(cluster_list):
#             f.write(' + {}{}{}'.format(cluster, self.sep, cluster))
#         # sum of dist > thr in the lower triangular part od the distance matrix
#
#         mask = (np.tril(distances, -1)>t)
#         threshold = np.tril(distances, -1).copy()
#         threshold[mask] = 0
#         s = np.sum(threshold) + 1
#         # s = np.sum(stats.threshold(np.tril(distances, -1), threshmax=t, newval=0)) + 1
#         logging.debug('ilp sum scores: '+str(s))
#         l = len(cluster_list)
#         for i in range(l):
#             cluster_i = cluster_list[i]
#             for j in range(i+1, l):
#                 if distances[i, j] < t:
#                     cluster_j = cluster_list[j]
#                     v = distances[i, j] / s
#                     f.write(
#                         ' +{} {}{}{} '.format(v, cluster_i, self.sep,
#                                                 cluster_j))
#         f.write("\nSubject to\n")
#
#         # x_i,i is a centre or not
#         for i, cluster_i in enumerate(cluster_list):
#             f.write('S{}: {}{}{}'.format(i, cluster_i, self.sep, cluster_i))
#             for j, cluster_j in enumerate(cluster_list):
#                 if i != j and distances[i, j] < t:
#                     f.write(" + {}{}{}".format(cluster_i, self.sep, cluster_j))
#             f.write(' = 1\n')
#
#         # center
#         for i, cluster_i in enumerate(cluster_list):
#             for j, cluster_j in enumerate(cluster_list):
#                 if i != j and distances[i, j] < t:
#                     f.write("{}{}{} - {}{}{} <=0\n".format(cluster_i, self.sep, cluster_j,
#                                                            cluster_j, self.sep, cluster_j))
#         f.write('End')
#
#     def _ilp_read(self, f):
#         """
#
#         :param f:
#         :return:
#         """
#         cluster_dict = dict()
#         for line in f:
#             line.strip()
#             if line.find(self.sep) != -1:
#                 lst = line.split()
#                 if len(lst) < 3:
#                     line += f.read()
#                     lst = line.split()
#                 names = lst[1].split(self.sep)
#                 if lst[3] == '1' and names[0] != names[1]:
#                     logging.debug('merge %s in %s', names[0], names[1])
#                     if names[1] not in cluster_dict:
#                         cluster_dict[names[1]] = []
#                     cluster_dict[names[1]].append(names[0])
#         return cluster_dict
#
#
# def ilp_iv(diar, scores, threshold=0.0):
#     """
#
#     :param diar:
#     :param scores:
#     :param threshold:
#     :return:
#     """
#     ilp = ILP_IV(diar, scores, threshold)
#     return ilp.perform()



# -*- coding: utf-8 -*-
#
# This file is part of s4d.
#
# s4d is a python package for speaker diarization.
# Home page: http://www-lium.univ-lemans.fr/s4d/
#
# s4d is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# s4d is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with s4d.  If not, see <http://www.gnu.org/licenses/>.


"""
Copyright 2014-2020 Sylvain Meignier and Anthony Larcher
"""

import multiprocessing
import numpy as np
import sys

from sidekit.frontend.features import *
from sidekit.frontend.vad import *
from sidekit.frontend.io import *
from sidekit.frontend.normfeat import *
from sidekit.sidekit_wrappers import *
from sidekit.features_server import FeaturesServer

if sys.version_info.major == 3:
    import queue as Queue
else:
    import Queue

__license__ = "LGPL"
__author__ = "Anthony Larcher"
__copyright__ = "Copyright 2014-2016 Anthony Larcher"
__maintainer__ = "Anthony Larcher"
__email__ = "anthony.larcher@univ-lemans.fr"
__status__ = "Production"
__docformat__ = 'reStructuredText'


class FeaturesServer_test(FeaturesServer):
    """
    A class for acoustic feature management.
    FeaturesServer should be used to extract acoustic features (MFCC or LFCC)
    from audio files in SPHERE, WAV or RAW PCM format.
    It can also be used to read and write acoustic features from and to disk
    in SPRO4 or HTK format.

    :attr input_dir: directory where to load audio or feature files
    :attr input_file_extension: extension of the incoming files
    :attrlabel_dir: directory where to read and write label files
    :attr label_files_extension: extension of label files to read and write
    :attr from_file: format of the input files to read, can be `audio`, `spro4`
        or `htk`, for audio files, format is given by the extension
    :attr config: pre-defined configuration for speaker diarization or recognition
        in 8 or 16kHz. Default is speaker recognition 8kHz
    :attr single_channel_extension: list with a single extension to add to 
        the audio filename when processing a single channel file. 
        Default is empty, means the feature file has the same speaker as
        the audio file
    :attr double_channel_extension: list of two channel extension to add 
        to the audio filename when processing two channel files. 
        Default is ['_a', '_b']
    :attr sampling_frequency: sample frequency in Hz, default is None, 
        determine when reading the audio file
    :attr lower_frequency: lower frequency limit of the filter bank
    :attr higher_frequency: higher frequency limit of the filter bank
    :attr linear_filters: number of linear filters to use for LFCC extraction
    :attr log_filters: number of linear filters to use for MFCC extraction
    :attr window_size: size of the sliding window in seconds
    :attr shift: time shift between two feature vectors
    :attr ceps_number: number of cepstral coefficients to extract
    :attr snr: snr level to consider for SNR-based voice activity detection
    :attr vad: type of voice activity detection to use, can be 'snr', 'energy' 
        (using a three Gaussian detector) or 'label' when reading the info from 
        pre-computed label files
    :attr feat_norm: normalization of the acoustic features, can be 
        'cms' for cepstral mean subtraction, 'mvn' for mean variance 
        normalization or 'stg' for short term Gaussianization
    :attr log_e: boolean, keep log energy
    :attr delta: boolean, add the first derivative of the cepstral coefficients
    :attr double_delta: boolean, add the second derivative of the cepstral 
        coefficients
    :attr rasta: boolean, perform RASTA filtering
    :attr keep_all_features: boolean, if False, only features labeled as 
        "speech" by the vad are saved if True, all features are saved and 
        a label file is produced

    """

    def __init__(self, input_dir=None,
                 feature_id=None,
                 config=None,
                 sampling_frequency=None,
                 lower_frequency=None,
                 higher_frequency=None,
                 linear_filters=None,
                 log_filters=None,
                 window_size=None,
                 shift=None,
                 ceps_number=None,
                 snr=None,
                 vad=None,
                 feat_norm=None,
                 log_e=None,
                 dct_pca=False,
                 dct_pca_config=None,
                 sdc=False,
                 sdc_config=None,
                 delta=None,
                 double_delta=None,
                 delta_filter=None,
                 rasta=None,
                 keep_all_features=None,
                 spec=False,
                 mspec=False,
                 mask=None
                 ):
        """ Process of extracting the feature frames (LFCC or MFCC) from an audio signal.
        Speech Activity Detection, MFCC (or LFCC) extraction and normalization.
        Can include RASTA filtering, Short Term Gaussianization, MVN and delta
        computation.

        :param input_dir: directory where to find the audio files.
                Default is ./
        :param input_file_extension: extension of the audio files to read.
                Default is 'sph'.
        :param label_dir: directory where to store label files is required.
                Default is ./
        :param label_file_extension: extension of the label files to create.
                Default is '.lbl'.
        :param configuration file : 'diar_16k', 'sid_16k', 'diar_8k' or 'sid_8k'
        """

        self.input_dir = './'
        self.from_file = 'audio'
        self.feature_id = 'ceps'
        self.sampling_frequency = 8000
        self.lower_frequency = 0
        self.higher_frequency = self.sampling_frequency / 2.
        self.linear_filters = 0
        self.log_filters = 40
        self.window_size = 0.025
        self.shift = 0.01
        self.ceps_number = 13
        self.snr = 40
        self.vad = None
        self.feat_norm = None
        self.log_e = False
        self.dct_pca = False        
        self.dct_pca_config = (12, 12, None)
        self.sdc = False
        self.sdc_config = (1, 3, 7)
        self.delta = False
        self.double_delta = False
        self.delta_filter = np.array([.25, .5, .25, 0, -.25, -.5, -.25])
        self.mask = None
        self.rasta = False
        self.keep_all_features = False
        self.spec = False
        self.mspec = False
        self.single_channel_extension = ['']
        self.double_channel_extension = ['_a', '_b']

        # If a predefined config is chosen, apply it
        if config == 'diar_16k':
            self._config_diar_16k()
        elif config == 'diar_8k':
            self._config_diar_8k()
        elif config == 'sid_8k':
            self._config_sid_8k()
        elif config == 'sid_16k':
            self._config_sid_16k()
        elif config == 'fb_8k':
            self._config_fb_8k()
        elif config is None:
            pass
        else:
            raise Exception('unknown configuration value')


        # Manually entered parameters are applied
        if input_dir is not None:
            self.input_dir = input_dir
        if feature_id is not None:
            self.feature_id = feature_id
        if sampling_frequency is not None:
            self.sampling_frequency = sampling_frequency
        if lower_frequency is not None:
            self.lower_frequency = lower_frequency
        if higher_frequency is not None:
            self.higher_frequency = higher_frequency
        if linear_filters is not None:
            self.linear_filters = linear_filters
        if log_filters is not None:
            self.log_filters = log_filters
        if window_size is not None:
            self.window_size = window_size
        if shift is not None:
            self.shift = shift
        if ceps_number is not None:
            self.ceps_number = ceps_number
        if snr is not None:
            self.snr = snr
        if vad is not None:
            self.vad = vad
        if feat_norm is not None:
            self.feat_norm = feat_norm
        if log_e is not None:
            self.log_e = log_e
        if dct_pca is not None:
            self.dct_pca = dct_pca
        if dct_pca_config is not None:
            self.dct_pca_config = dct_pca_config
        if sdc is not None:
            self.sdc = sdc
        if sdc_config is not None:
            self.sdc_config = sdc_config
        if delta is not None:
            self.delta = delta
        if double_delta is not None:
            self.double_delta = double_delta
        if delta_filter is not None:
            self.delta_filter = delta_filter
        if mask is not None:
            self.mask = mask
        if rasta is not None:
            self.rasta = rasta
        if keep_all_features is not None:
            self.keep_all_features = keep_all_features
        if spec:
            self.spec = True
        if mspec:
            self.mspec = True
        
        self.cep = []
        self.label = []
        self.show = 'empty'
        self.audio_filename = 'empty'
        root, ext = os.path.splitext(self.input_dir)
        if ext == '.hdf5' or ext == '.h5':
            self.from_file = 'hdf5'

    def __repr__(self):
        ch = '\t show: {} keep_all_features: {} from_file: {}\n'.format(
            self.show, self.keep_all_features, self.from_file)
        ch += '\t inputDir: {}  \n'.format(self.input_dir)
        ch += '\t lower_frequency: {}  higher_frequency: {} \n'.format(
            self.lower_frequency, self.higher_frequency)
        ch += '\t sampling_frequency: {} '.format(self.sampling_frequency)
        ch += '\t linear_filters: {}  or log_filters: {} \n'.format(
            self.linear_filters, self.log_filters)
        ch += '\t ceps_number: {}  window_size: {} shift: {} \n'.format(
            self.ceps_number, self.window_size, self.shift)
        ch += '\t vad: {}  snr: {} \n'.format(self.vad, self.snr)
        ch += '\t feat_norm: {} rasta: {} \n'.format(self.feat_norm, self.rasta)
        ch += '\t log_e: {} delta: {} double_delta: {} \n'.format(self.log_e,
                                                                  self.delta,
                                                                  self.double_delta)
        return ch

    def _config_diar_16k(self):
        """
        12 MFCC + E, no normalization
        """
        self.sampling_frequency = 16000
        self.lower_frequency = 133.3333
        self.higher_frequency = 6855.4976
        self.linear_filters = 0
        self.log_filters = 40
        self.window_size = 0.025
        self.shift = 0.01
        self.ceps_number = 13
        self.snr = 40
        self.vad = None
        self.feat_norm = None
        self.log_e = True
        self.delta = False
        self.double_delta = False
        self.rasta = False
        self.keep_all_features = True

    def _config_diar_8k(self):
        """
        12 MFCC + E, no normalization
        """
        self.sampling_frequency = 8000
        self.lower_frequency = None
        self.higher_frequency = None
        self.linear_filters = 0
        self.log_filters = 24
        self.window_size = 0.025
        self.shift = 0.01
        self.ceps_number = 13
        self.snr = 40
        self.vad = None
        self.feat_norm = None
        self.log_e = True
        self.delta = False
        self.double_delta = False
        self.rasta = False
        self.keep_all_features = True

    def _config_sid_16k(self):
        """
        19 MFCC + E + D + DD, normalization cmvn
        """
        self.sampling_frequency = 16000
        self.lower_frequency = 133.3333
        self.higher_frequency = 6855.4976
        self.linear_filters = 0
        self.log_filters = 40
        self.window_size = 0.025
        self.shift = 0.01
        self.ceps_number = 13
        self.snr = 40
        self.vad = 'snr'
        self.feat_norm = 'cmvn'
        self.log_e = True
        self.delta = True
        self.double_delta = True
        self.rasta = True
        self.keep_all_features = False

    def _config_sid_8k(self):
        """
        19 MFCC + E + D + DD, normalization cmvn
        """
        self.sampling_frequency = 8000
        self.lower_frequency = 200
        self.higher_frequency = 3800
        self.linear_filters = 0
        self.log_filters = 24
        self.window_size = 0.025
        self.shift = 0.01
        self.ceps_number = 13
        self.snr = 40
        self.vad = 'snr'
        self.feat_norm = 'cmvn'
        self.log_e = True
        self.delta = True
        self.double_delta = True
        self.rasta = True
        self.keep_all_features = False

    def _config_fb_8k(self):
        """
        19 MFCC + E + D + DD, normalization cmvn
        """
        self.sampling_frequency = 8000
        self.lower_frequency = 300
        self.higher_frequency = 3400
        self.linear_filters = 0
        self.log_filters = 40
        self.window_size = 0.025
        self.shift = 0.01
        self.ceps_number = 0
        self.snr = 40
        self.vad = None
        self.feat_norm = None
        self.log_e = False
        self.delta = False
        self.double_delta = False
        self.rasta = False
        self.keep_all_features = True
        self.mspec = True
 
    def _config_lid_8k_sdc(self):
        """
        7 MFCC + 1 - 3 - 7 SDC
        """
        self.sampling_frequency = 8000
        self.lower_frequency = 300
        self.higher_frequency = 3400
        self.linear_filters = 0
        self.log_filters = 24
        self.window_size = 0.025
        self.shift = 0.01
        self.ceps_number = 7
        self.snr = 40
        self.vad = 'snr'
        self.feat_norm = None
        self.log_e = False
        self.delta = False
        self.double_delta = False
        self.sdc = True
        self.sdc_config = (1, 3, 7)
        self.rasta = False
        self.keep_all_features = False

    def _features(self, show):
        cep = None
        label = None
        window_sample = int(self.window_size * self.sampling_frequency)
        shift_sample = int(self.shift * self.sampling_frequency)

        audio_filename = self.input_dir.format(s=show)
        logging.debug('--> '+audio_filename)
        if not os.path.isfile(audio_filename):
            logging.error('%s %s', audio_filename, show)
            raise IOError('File ' + audio_filename + ' not found')
        logging.info('read audio')
        x, rate = read_audio(audio_filename, self.sampling_frequency)

        if rate != self.sampling_frequency:
            raise "file rate don't match the rate of the feature server configuration"
        self.audio_filename = audio_filename
        logging.info('size of signal: %f len %d type size %d', x.nbytes/1024/1024, len(x), x.nbytes/len(x))

        if x.ndim == 1:
            x = x[:, np.newaxis]

        for i in range(0, 200, 5):
            print('==> ', i, x[i:i+5])
        
        channel_ext = []
        channel_nb = x.shape[1]
        np.random.seed(0)
        #x[:, 0] += 0.0001 * numpy.random.randn(x.shape[0])

        if channel_nb == 1:
            channel_ext.append('')
            # Random noise is added to the input signal to avoid zero frames.
        elif channel_nb == 2:
            channel_ext.append('_a')
            channel_ext.append('_b')
            #x[:, 1] += 0.0001 * numpy.random.randn(x.shape[0])

        # Process channels one by one
        for chan, chan_ext in enumerate(channel_ext):
            l = x.shape[0]

            dec = shift_sample * 250 * 25000 + window_sample
            dec2 = window_sample - shift_sample
            start = 0
            end = min(dec, l)
            while start < l - dec2:
                # if end < l:
                logging.info('process part : %f %f %f',
                             start / self.sampling_frequency,
                             end / self.sampling_frequency,
                             l / self.sampling_frequency)

                tmp = self._features_chan(show, channel_ext, x[start:end, chan])

                if cep is None:
                    cep = []
                    label = []
                    cep.append(tmp[0])
                    label.append(tmp[1])
                else:
                    cep.append(tmp[0])
                    label.append(tmp[1])
                start = end - dec2
                end = min(end + dec, l)
                if cep[-1].shape[0] > 0:
                    logging.info('!! size of signal cep: %f len %d type size %d', cep[-1].nbytes/1024/1024, len(cep[-1]),
                             cep[-1].nbytes/len(cep[-1]))
        del x
        # Smooth the cluster_list and fuse the channels if more than one.
        logging.info('Smooth the cluster_list and fuse the channels if more than one')
        if self.vad is not None:
            label = label_fusion(label)
        self._normalize(label, cep)

        # Keep only the required features and save the appropriate files
        # which are either feature files alone or feature and label files
        if not self.keep_all_features:
            logging.info('no keep all')
            for chan, chan_ext in enumerate(channel_ext):
                cep[chan] = cep[chan][label[chan]]
                label[chan] = label[chan][label[chan]]

        return cep, label

    def _features_chan(self, show, channel_ext, x):

        """Compelete the overwhole process of extracting the feature frames
        (LFCC or MFCC) from an audio signal.
        Speech Activity Detection, MFCC (or LFCC) extraction and normalization.
        Can include RASTA filtering, Short Term Gaussianization, MVN and delta
        computation.

        :param show: speaker of the file.
        """
        # If the size of the signal is not enough for one frame, return zero features
        if x.shape[0] < self.sampling_frequency * self.window_size:
            cep_size = self.ceps_number * (1 + int(self.delta) + int(self.double_delta))\
                       + int(self.mspec) * (self.linear_filters + self.log_filters)
            cep = np.empty((0, cep_size))
            label = np.empty((0, 1))

        # Extract cepstral coefficients
        else:
            c = mfcc(x, fs=self.sampling_frequency,
                 lowfreq=self.lower_frequency,
                 maxfreq=self.higher_frequency,
                 nlinfilt=self.linear_filters,
                 nwin=self.window_size, nlogfilt=self.log_filters,
                 nceps=self.ceps_number, get_spec=self.spec, 
                 get_mspec=self.mspec)

            if self.ceps_number == 0 and self.mspec:
                cep = c[3]
                label = self._vad(c[1], x, channel_ext, show)

            else:
                label = self._vad(c[1], x, channel_ext, show)

                cep = self._log_e(c)
                cep, label = self._rasta(cep, label)
                if self.delta or self.double_delta:
                    cep = self._delta_and_2delta(cep)
                elif self.dct_pca:
                    cep = pca_dct(cep, self.dct_pca_config[0],
                                  self.dct_pca_config[1],
                                  self.dct_pca_config[2])
                elif self.sdc:
                    cep = shifted_delta_cepstral(cep, d=self.sdc_config[0],
                                                 P=self.sdc_config[1],
                                                 k=self.sdc_config[2])
        return cep, label

    def _log_e(self, c):
        """If required, add the log energy as last coefficient"""
        if self.log_e:
            logging.info('keep log_e')
            return np.hstack((c[1][:, np.newaxis], c[0]))
        else:
            logging.info('don\'t keep c0')
            return c[0]

    def _vad(self, logEnergy, x, channel_ext, show):
        """
        Apply Voice Activity Detection.
        :param x:
        :param channel:
        :param window_sample:
        :param channel_ext:
        :param show:
        :return:
        """
        label = None
        if self.vad is None:
            logging.info('no vad')
            label = np.array([True] * logEnergy.shape[0])
        elif self.vad == 'snr':
            logging.info('vad : snr')
            window_sample = int(self.window_size * self.sampling_frequency)
            label = vad_snr(x, self.snr, fs=self.sampling_frequency,
                            shift=self.shift, nwin=window_sample)
        elif self.vad == 'energy':
            logging.info('vad : energy')
            label = vad_energy(logEnergy, distribNb=3,
                               nbTrainIt=8, flooring=0.0001,
                               ceiling=1.5, alpha=0.1)
        else:
            logging.warning('Wrong VAD type')
        return label

    def _rasta(self, cep, label):
        """
        Performs RASTA filtering if required.
        The two first frames are copied from the third to keep
        the length consistent
        !!! if vad is None: label[] is empty

        :param channel: speaker of the channel
        :return:
        """
        if self.rasta:
            logging.info('perform RASTA %s', self.rasta)
            cep = rasta_filt(cep)
            cep[:2, :] = cep[2, :]
            label[:2] = label[2]
            
        return cep, label

    def _delta_and_2delta(self, cep):
        """
        Add deltas and double deltas.
        :param cep: a matrix of cepstral cefficients
        
        :return: the cepstral coefficient stacked with deltas and double deltas
        """
        if self.delta:
            logging.info('add delta')
            delta = compute_delta(cep, filt=self.delta_filter)
            cep = np.column_stack((cep, delta))
            if self.double_delta:
                logging.info('add delta delta')
                double_delta = compute_delta(delta, filt=self.delta_filter)
                cep = np.column_stack((cep, double_delta))
        return cep

    def _normalize(self, label, cep):
        """
        Normalize features in place

        :param label:
        :return:
        """
        # Perform feature normalization on the entire session.
        if self.feat_norm is None:
            logging.info('no normalization')
            pass
        elif self.feat_norm == 'cms':
            logging.info('cms normalization')
            for chan, c in enumerate(cep):
                cms(cep[chan], label[chan])
        elif self.feat_norm == 'cmvn':
            logging.info('cmvn normalization')
            for chan, c in enumerate(cep):
                cmvn(cep[chan], label[chan])
        elif self.feat_norm == 'stg':
            logging.info('stg normalization')
            for chan, c in enumerate(cep):
                stg(cep[chan], label=label[chan])
        elif self.feat_norm == 'cmvn_sliding':
            logging.info('sliding cmvn normalization')
            for chan, c in enumerate(cep):
                cep_sliding_norm(cep[chan], win=301, center=True, reduce=True)
        elif self.feat_norm == 'cms_sliding':
            logging.info('sliding cms normalization')
            for chan, c in enumerate(cep):
                cep_sliding_norm(cep[chan], win=301, center=True, reduce=False)
        else:
            logging.warning('Wrong feature normalisation type')

    def load(self, show, id=None):
        """
        Load a cep from audio or mfcc file. This method loads all channels
        available in the file.
        
        :param show: the speaker of the show to load
        
        :return: the cep array and the label array
        """
        # test if features is already computed
        if self.show == show:
            return self.cep, self.label
        self.show = show
        if self.from_file == 'audio':
            logging.debug('compute MFCC: ' + show)
            logging.debug(self.__repr__())
            self.cep, self.label = self._features(show)
        elif self.from_file == 'hdf5':
            logging.debug('load hdf5: ' + show)
            input_filename = self.input_dir.format(s=show)
            with h5py.File(input_filename, "r") as hdf5_input_fh:
                logging.debug('*** '+input_filename+' '+show)
                vad = True
                if self.vad is None:
                    vad = False
                cep, label = read_hdf5(hdf5_input_fh, show, feature_id=self.feature_id, label=vad)
                self.cep = [cep]
                if label is None:
                    self.label = [np.array([True] * self.cep[0].shape[0])]
                else:
                    self.label = [label]
        else:
            raise Exception('unknown from_file value')

        if self.mask is not None:
            self.cep[0] = self._mask(self.cep[0])

        if not self.keep_all_features:
            logging.debug('!!! no keep all feature !!!')
            self.cep[0] = self.cep[0][self.label[0]]
            self.label[0] = [np.array([True] * self.cep[0].shape[0])]

        return self.cep, self.label

    def _mask(self, cep):
        """
        keep only the MFCC index present in the filter list
        :param cep:
        :return: return the list of MFCC given by filter list
        """
        if len(self.mask) == 0:
            raise Exception('filter list is empty')
        logging.debug('applied mask')
        return cep[:, self.mask]

    def save(self, show, filename, mfcc_format, and_label=True):
        """
        Save the cep array in file
        
        :param show: the speaker of the show to save (loaded if need)
        :param filename: the file speaker of the mffc file or a list of 2 filenames
            for the case of double channel files
        :param mfcc_format: format of the mfcc file taken in values
            ['pickle', 'spro4', 'htk']
        :param and_label: boolean, if True save label files
        
        :raise: Exception if feature format is unknown
        """
        self.load(show)

        hdf5_ouput_fh = h5py.File(filename, "w")
        logging.debug('save hdf5: ' + show)
        #write_hdf5(show, fh, feat, feat_type='ceps', label=None )
        write_hdf5(show, hdf5_ouput_fh, self.cep[0], label=self.label[0])
        hdf5_ouput_fh.close()


    @process_parallel_lists
    def save_list(self, audio_file_list, feature_file_list, mfcc_format, feature_dir, 
                  feature_file_extension, and_label=False, numThread=1):
        """
        Function that takes a list of audio files and extract features
        
        :param audio_file_list: an array of string containing the speaker of the feature
            files to load
        :param feature_file_list: list of feature files to save, should correspond to the input audio_file_list
        :param mfcc_format: format of the feature files to save, could be spro4, htk, pickle
        :param feature_dir: directory where to save the feature files
        :param feature_file_extension: extension of the feature files to save
        :param and_label: boolean, if True save the label files
        :param numThread: number of parallel process to run
        """
        logging.info(self)
        for audio_file, feature_file in zip(audio_file_list, feature_file_list):
            cep_filename = os.path.join(feature_dir, feature_file + feature_file_extension)
            self.save(audio_file, cep_filename, mfcc_format, and_label)

    def dim(self):
        if self.show != 'empty':
            return self.cep[0].shape[1]
        dim = self.ceps_number
        if self.log_e:
            dim += 1
        if self.delta:
            dim *= 2
        if self.double_delta:
            dim *= 2
        logging.warning('cep dim computed using featureServer parameters')
        return dim

    def save_parallel(self, input_audio_list, output_feature_list, mfcc_format, feature_dir,
                      feature_file_extension, and_label=False, numThread=1):
        """
        Extract features from audio file using parallel computation
        
        :param input_audio_list: an array of string containing the speaker
            of the audio files to process
        :param output_feature_list: an array of string containing the 
            speaker of the features files to save
        :param mfcc_format: format of the output feature files, could be spro4, htk, pickle
        :param feature_dir: directory where to save the feature files
        :param feature_file_extension: extension of the feature files to save
        :param and_label: boolean, if True save the label files
        :param numThread: number of parallel process to run
        """
        # Split the features to process for multi-threading
        loa = np.array_split(input_audio_list, numThread)
        lof = np.array_split(output_feature_list, numThread)
    
        jobs = []
        multiprocessing.freeze_support()
        for idx, feat in enumerate(loa):
            p = multiprocessing.Process(target=self.save_list,
                                        args=(loa[idx], lof[idx], mfcc_format, feature_dir,
                                              feature_file_extension, and_label))
            jobs.append(p)
            p.start()
        for p in jobs:
            p.join()

    def _load_and_stack_worker(self, input_queue, output):
        """Load a list of feature files into a Queue object
        
        :param input_queue: a Queue object
        :param output: a list of Queue objects to fill
        """
        while True:
            next_task = input_queue.get()
            
            if next_task is None:
                # Poison pill means shutdown
                output.put(None)
                input_queue.task_done()
                break
            
            # check which channel to keep from the file
            if next_task.endswith(self.double_channel_extension[0]) and (self.from_file == 'audio'):
                next_task = next_task[:-len(self.double_channel_extension[0])]
                output.put(self.load(next_task)[0][0])
            if next_task.endswith(self.double_channel_extension[1]) and self.from_file == 'audio':
                next_task = next_task[:-len(self.double_channel_extension[1])]
                output.put(self.load(next_task)[0][1])
            else:
                cep = self.load(next_task)[0][0]
                output.put(cep)
            
            input_queue.task_done()

    def load_and_stack(self, fileList, numThread=1):
        """Load a list of feature files and stack them in a unique ndarray. 
        The list of files to load is splited in sublists processed in parallel
        
        :param fileList: a list of files to load
        :param numThread: numbe of thead (optional, default is 1)
        """
        queue_in = multiprocessing.JoinableQueue(maxsize=len(fileList)+numThread)
        queue_out = []
        
        # Start worker processes
        jobs = []
        for i in range(numThread):
            queue_out.append(multiprocessing.Queue())
            p = multiprocessing.Process(target=self._load_and_stack_worker, 
                                        args=(queue_in, queue_out[i]))
            jobs.append(p)
            p.start()
        
        # Submit tasks
        for task in fileList:
            queue_in.put(task)

        for task in range(numThread):
            queue_in.put(None)
        
        # Wait for all the tasks to finish
        queue_in.join()
                   
        output = []
        for q in queue_out:
            while True:
                data = q.get()
                if data is None:
                    break
                output.append(data)

        for p in jobs:
            p.join()
        all_cep = np.concatenate(output, axis=0)

        return all_cep

    def load_and_stack_threading(self, fileList, numThread=1):
        """Load a list of feature files and stack them in a unique ndarray. 
        The list of files to load is splited in sublists processed in parallel
        
        :param fileList: a list of files to load
        :param numThread: numbe of thead (optional, default is 1)
        """
        queue_in = multiprocessing.JoinableQueue(maxsize=len(fileList)+numThread)
        queue_out = []
        
        # Start worker processes
        jobs = []
        for i in range(numThread):
            queue_out.append(Queue.Queue())
            p = threading.Thread(target=self._load_and_stack_worker, args=(queue_in, queue_out[i]))
            jobs.append(p)
            p.start()
        
        # Submit tasks
        for task in fileList:
            queue_in.put(task)

        for task in range(numThread):
            queue_in.put(None)
        
        # Wait for all the tasks to finish
        queue_in.join()
                   
        output = []
        for q in queue_out:
            while True:
                data = q.get()
                if data is None:
                    break
                output.append(data)

        for p in jobs:
            p.join()
        all_cep = np.concatenate(output, axis=0)

        return all_cep

    def mean_std(self, filename):
        feat = self.load(filename)[0][0]
        return feat.shape[0], feat.sum(axis=0), np.sum(feat**2, axis=0)



"""
Gaussian Mixture Models
"""

# Author: Ron Weiss <ronweiss@gmail.com>
#         Fabian Pedregosa <fabian.pedregosa@inria.fr>
#

import numpy as np

from sklearn.base import BaseEstimator
from sklearn import cluster


def logsum(A, axis=None):
    """Computes the sum of A assuming A is in the log domain.

    Returns log(sum(exp(A), axis)) while minimizing the possibility of
    over/underflow.
    """
    Amax = A.max(axis)
    if axis and A.ndim > 1:
        shape = list(A.shape)
        shape[axis] = 1
        Amax.shape = shape
    Asum = np.log(np.sum(np.exp(A - Amax), axis))
    Asum += Amax.reshape(Asum.shape)
    if axis:
        # Look out for underflow.
        Asum[np.isnan(Asum)] = - np.Inf
    return Asum


# TODO: this lacks a docstring
def normalize(A, axis=None):
    A += np.finfo(float).eps
    Asum = A.sum(axis)
    if axis and A.ndim > 1:
        # Make sure we don't divide by zero.
        Asum[Asum == 0] = 1
        shape = list(A.shape)
        shape[axis] = 1
        Asum.shape = shape
    return A / Asum


def lmvnpdf(obs, means, covars, cvtype='diag'):
    """Compute the log probability under a multivariate Gaussian distribution.

    Parameters
    ----------
    obs : array_like, shape (O, D)
        List of D-dimensional data points.  Each row corresponds to a
        single data point.
    means : array_like, shape (C, D)
        List of D-dimensional mean vectors for C Gaussians.  Each row
        corresponds to a single mean vector.
    covars : array_like
        List of C covariance parameters for each Gaussian.  The shape
        depends on `cvtype`:
            (C,)      if 'spherical',
            (D, D)    if 'tied',
            (C, D)    if 'diag',
            (C, D, D) if 'full'
    cvtype : string
        Type of the covariance parameters.  Must be one of
        'spherical', 'tied', 'diag', 'full'.  Defaults to 'diag'.

    Returns
    -------
    lpr : array_like, shape (O, C)
        Array containing the log probabilities of each data point in
        `observation` under each of the C multivariate Gaussian distributions.
    """
    lmvnpdf_dict = {'spherical': _lmvnpdfspherical,
                    'tied': _lmvnpdftied,
                    'diag': _lmvnpdfdiag,
                    'full': _lmvnpdffull}
    return lmvnpdf_dict[cvtype](obs, means, covars)


def sample_gaussian(mean, covar, cvtype='diag', n_samples=1):
    """Generate random samples from a Gaussian distribution.

    Parameters
    ----------
    mean : array_like, shape (n_features,)
        Mean of the distribution.

    covars : array_like, optional
        Covariance of the distribution.  The shape depends on `cvtype`:
            scalar  if 'spherical',
            (D)     if 'diag',
            (D, D)  if 'tied', or 'full'

    cvtype : string, optional
        Type of the covariance parameters.  Must be one of
        'spherical', 'tied', 'diag', 'full'.  Defaults to 'diag'.

    n_samples : int, optional
        Number of samples to generate. Defaults to 1.

    Returns
    -------
    observation : array, shape (n_features, n_samples)
        Randomly generated sample
    """
    n_dim = len(mean)
    rand = np.random.randn(n_dim, n_samples)
    if n_samples == 1:
        rand.shape = (n_dim,)

    if cvtype == 'spherical':
        rand *= np.sqrt(covar)
    elif cvtype == 'diag':
        rand = np.dot(np.diag(np.sqrt(covar)), rand)
    else:
        from scipy import linalg
        U, s, V = linalg.svd(covar)
        sqrtS = np.diag(np.sqrt(s))
        sqrt_covar = np.dot(U, np.dot(sqrtS, V))
        rand = np.dot(sqrt_covar, rand)

    return (rand.T + mean).T


class GMM(BaseEstimator):
    """Gaussian Mixture Model

    Representation of a Gaussian mixture model probability distribution.
    This class allows for easy evaluation of, sampling from, and
    maximum-likelihood estimation of the parameters of a GMM distribution.

    Initializes parameters such that every mixture component has zero
    mean and identity covariance.


    Parameters
    ----------
    n_states : int, optional
        Number of mixture components. Defaults to 1.

    cvtype : string (read-only), optional
        String describing the type of covariance parameters to
        use.  Must be one of 'spherical', 'tied', 'diag', 'full'.
        Defaults to 'diag'.


    Attributes
    ----------
    cvtype : string (read-only)
        String describing the type of covariance parameters used by
        the GMM.  Must be one of 'spherical', 'tied', 'diag', 'full'.
    n_features : int
        Dimensionality of the Gaussians.
    n_states : int (read-only)
        Number of mixture components.
    weights : array, shape (`n_states`,)
        Mixing weights for each mixture component.
    means : array, shape (`n_states`, `n_features`)
        Mean parameters for each mixture component.
    covars : array
        Covariance parameters for each mixture component.  The shape
        depends on `cvtype`:
            (`n_states`,)                             if 'spherical',
            (`n_features`, `n_features`)              if 'tied',
            (`n_states`, `n_features`)                if 'diag',
            (`n_states`, `n_features`, `n_features`)  if 'full'
    converged_ : bool
        True when convergence was reached in fit(), False
        otherwise.

    Methods
    -------
    decode(X)
        Find most likely mixture components for each point in `X`.
    eval(X)
        Compute the log likelihood of `X` under the model and the
        posterior distribution over mixture components.
    fit(X)
        Estimate model parameters from `X` using the EM algorithm.
    predict(X)
        Like decode, find most likely mixtures components for each
        observation in `X`.
    rvs(n=1)
        Generate `n` samples from the model.
    score(X)
        Compute the log likelihood of `X` under the model.

    Examples
    --------
    >>> import numpy as numpy
    >>> g = mixture.GMM(n_states=2)

    >>> # Generate random observations with two modes centered on 0
    >>> # and 10 to use for training.
    >>> numpy.random.seed(0)
    >>> observation = numpy.concatenate((numpy.random.randn(100, 1),
    ...                       10 + numpy.random.randn(300, 1)))
    >>> g.fit(observation)
    GMM(cvtype='diag', n_states=2)
    >>> g.weights
    array([ 0.25,  0.75])
    >>> g.means
    array([[ 0.05980802],
           [ 9.94199467]])
    >>> g.covars
    [array([[ 1.01682662]]), array([[ 0.96080513]])]
    >>> numpy.round(g.weights, 2)
    array([ 0.25,  0.75])
    >>> numpy.round(g.means, 2)
    array([[ 0.06],
           [ 9.94]])
    >>> numpy.round(g.covars, 2)
    ... #doctest: +NORMALIZE_WHITESPACE
    array([[[ 1.02]],
           [[ 0.96]]])
    >>> g.predict([[0], [2], [9], [10]])
    array([0, 0, 1, 1])
    >>> numpy.round(g.score([[0], [2], [9], [10]]), 2)
    array([-2.32, -4.16, -1.65, -1.19])

    >>> # Refit the model on new data (initial parameters remain the
    >>> # same), this time with an even split between the two modes.
    >>> g.fit(20 * [[0]] +  20 * [[10]])
    GMM(cvtype='diag', n_states=2)
    >>> numpy.round(g.weights, 2)
    array([ 0.5,  0.5])
    """

    def __init__(self, n_states=1, cvtype='diag'):
        self._n_states = n_states
        self._cvtype = cvtype

        if not cvtype in ['spherical', 'tied', 'diag', 'full']:
            raise ValueError('bad cvtype')

        self.weights = np.ones(self._n_states) / self._n_states

        # flag to indicate exit status of fit() method: converged (True) or
        # n_iter reached (False)
        self.converged_ = False

    # Read-only properties.
    @property
    def cvtype(self):
        """Covariance type of the model.

        Must be one of 'spherical', 'tied', 'diag', 'full'.
        """
        return self._cvtype

    @property
    def n_states(self):
        """Number of mixture components in the model."""
        return self._n_states

    def _get_covars(self):
        """Return covars as a full matrix."""
        if self.cvtype == 'full':
            return self._covars
        elif self.cvtype == 'diag':
            return [np.diag(cov) for cov in self._covars]
        elif self.cvtype == 'tied':
            return [self._covars] * self._n_states
        elif self.cvtype == 'spherical':
            return [np.eye(self.n_features) * f for f in self._covars]

    def _set_covars(self, covars):
        covars = np.asanyarray(covars)
        _validate_covars(covars, self._cvtype, self._n_states, self.n_features)
        self._covars = covars

    covars = property(_get_covars, _set_covars)

    def _get_means(self):
        """Mean parameters for each mixture component."""
        return self._means

    def _set_means(self, means):
        means = np.asarray(means)
        if hasattr(self, 'n_features') and \
               means.shape != (self._n_states, self.n_features):
            raise ValueError('means must have shape (n_states, n_features)')
        self._means = means.copy()
        self.n_features = self._means.shape[1]

    means = property(_get_means, _set_means)

    def _get_weights(self):
        """Mixing weights for each mixture component."""
        return np.exp(self._log_weights)

    def _set_weights(self, weights):
        if len(weights) != self._n_states:
            raise ValueError('weights must have length n_states')
        if not np.allclose(np.sum(weights), 1.0):
            raise ValueError('weights must sum to 1.0')

        self._log_weights = np.log(np.asarray(weights).copy())

    weights = property(_get_weights, _set_weights)

    def eval(self, obs):
        """Evaluate the model on data

        Compute the log probability of `observation` under the model and
        return the posterior distribution (responsibilities) of each
        mixture component for each element of `observation`.

        Parameters
        ----------
        obs : array_like, shape (n_samples, n_features)
            List of n_features-dimensional data points.  Each row
            corresponds to a single data point.

        Returns
        -------
        logprob : array_like, shape (n_samples,)
            Log probabilities of each data point in `observation`
        posteriors: array_like, shape (n_samples, n_states)
            Posterior probabilities of each mixture component for each
            observation
        """
        obs = np.asanyarray(obs)
        lpr = (lmvnpdf(obs, self._means, self._covars, self._cvtype)
               + self._log_weights)
        logprob = logsum(lpr, axis=1)
        posteriors = np.exp(lpr - logprob[:, np.newaxis])
        return logprob, posteriors

    def score(self, obs):
        """Compute the log probability under the model.

        Parameters
        ----------
        obs : array_like, shape (n_samples, n_features)
            List of n_features-dimensional data points.  Each row
            corresponds to a single data point.

        Returns
        -------
        logprob : array_like, shape (n_samples,)
            Log probabilities of each data point in `observation`
        """
        logprob, posteriors = self.eval(obs)
        return logprob

    def decode(self, obs):
        """Find most likely mixture components for each point in `observation`.

        Parameters
        ----------
        obs : array_like, shape (n, n_features)
            List of n_features-dimensional data points.  Each row
            corresponds to a single data point.

        Returns
        -------
        logprobs : array_like, shape (n_samples,)
            Log probability of each point in `observation` under the model.
        components : array_like, shape (n_samples,)
            Index of the most likelihod mixture components for each observation
        """
        logprob, posteriors = self.eval(obs)
        return logprob, posteriors.argmax(axis=1)

    def predict(self, X):
        """Predict label for data.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]

        Returns
        -------
        C : array, shape = (n_samples,)
        """
        logprob, components = self.decode(X)
        return components

    def predict_proba(self, X):
        """Predict posterior probability of data under each Gaussian
        in the model.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]

        Returns
        -------
        T : array-like, shape = (n_samples, n_states)
            Returns the probability of the sample for each Gaussian
            (state) in the model.
        """
        logprob, posteriors = self.eval(X)
        return posteriors

    def rvs(self, n_samples=1):
        """Generate random samples from the model.

        Parameters
        ----------
        n_samples : int, optional
            Number of samples to generate. Defaults to 1.

        Returns
        -------
        observation : array_like, shape (n_samples, n_features)
            List of samples
        """
        weight_pdf = self.weights
        weight_cdf = np.cumsum(weight_pdf)

        obs = np.empty((n_samples, self.n_features))
        rand = np.random.rand(n_samples)
        # decide which component to use for each sample
        comps = weight_cdf.searchsorted(rand)
        # for each component, generate all needed samples
        for comp in range(self._n_states):
            # occurrences of current component in observation
            comp_in_obs = (comp == comps)
            # number of those occurrences
            num_comp_in_obs = comp_in_obs.sum()
            if num_comp_in_obs > 0:
                if self._cvtype == 'tied':
                    cv = self._covars
                else:
                    cv = self._covars[comp]
                obs[comp_in_obs] = sample_gaussian(
                    self._means[comp], cv, self._cvtype, num_comp_in_obs).T
        return obs

    def fit(self, X, n_iter=10, min_covar=1e-3, thresh=1e-2, params='wmc',
            init_params='wmc'):
        """Estimate model parameters with the expectation-maximization
        algorithm.

        A initialization step is performed before entering the em
        algorithm. If you want to avoid this step, set the keyword
        argument init_params to the empty string ''. Likewise, if you
        would like just to do an initialization, call this method with
        n_iter=0.

        Parameters
        ----------
        X : array_like, shape (n, n_features)
            List of n_features-dimensional data points.  Each row
            corresponds to a single data point.
        n_iter : int, optional
            Number of EM iterations to perform.
        min_covar : float, optional
            Floor on the diagonal of the covariance matrix to prevent
            overfitting.  Defaults to 1e-3.
        thresh : float, optional
            Convergence threshold.
        params : string, optional
            Controls which parameters are updated in the training
            process.  Can contain any combination of 'w' for weights,
            'm' for means, and 'c' for covars.  Defaults to 'wmc'.
        init_params : string, optional
            Controls which parameters are updated in the initialization
            process.  Can contain any combination of 'w' for weights,
            'm' for means, and 'c' for covars.  Defaults to 'wmc'.
        """

        ## initialization step

        X = np.asanyarray(X)

        if hasattr(self, 'n_features') and self.n_features != X.shape[1]:
            raise ValueError('Unexpected number of dimensions, got %s but '
                             'expected %s' % (X.shape[1], self.n_features))

        self.n_features = X.shape[1]

        if 'm' in init_params:
            self._means = cluster.KMeans(
                k=self._n_states).fit(X).cluster_centers_
        elif not hasattr(self, 'means'):
                self._means = np.zeros((self.n_states, self.n_features))

        if 'w' in init_params or not hasattr(self, 'weights'):
            self.weights = np.tile(1.0 / self._n_states, self._n_states)

        if 'c' in init_params:
            cv = np.cov(X.T)
            if not cv.shape:
                cv.shape = (1, 1)
            self._covars = _distribute_covar_matrix_to_match_cvtype(
                cv, self._cvtype, self._n_states)
        elif not hasattr(self, 'covars'):
                self.covars = _distribute_covar_matrix_to_match_cvtype(
                    np.eye(self.n_features), self.cvtype, self.n_states)

        # EM algorithm
        logprob = []
        # reset self.converged_ to False
        self.converged_ = False
        for i in range(n_iter):
            # Expectation step
            curr_logprob, posteriors = self.eval(X)
            logprob.append(curr_logprob.sum())

            # Check for convergence.
            if i > 0 and abs(logprob[-1] - logprob[-2]) < thresh:
                self.converged_ = True
                break

            # Maximization step
            self._do_mstep(X, posteriors, params, min_covar)

        return self

    def _do_mstep(self, X, posteriors, params, min_covar=0):
        w = posteriors.sum(axis=0)
        avg_obs = np.dot(posteriors.T, X)
        norm = 1.0 / (w[:, np.newaxis] + 1e-200)

        if 'w' in params:
            self._log_weights = np.log(w / w.sum())
        if 'm' in params:
            self._means = avg_obs * norm
        if 'c' in params:
            covar_mstep_func = _covar_mstep_funcs[self._cvtype]
            self._covars = covar_mstep_func(self, X, posteriors,
                                            avg_obs, norm, min_covar)

        return w


##
## some helper routines
##


def _lmvnpdfdiag(obs, means=0.0, covars=1.0):
    n_obs, n_dim = obs.shape
    # (x-y).T A (x-y) = x.T A x - 2x.T A y + y.T A y
    #lpr = -0.5 * (numpy.tile((numpy.sum((means**2) / covars, 1)
    #                  + numpy.sum(numpy.log(covars), 1))[numpy.newaxis,:], (n_obs,1))
    lpr = -0.5 * (n_dim * np.log(2 * np.pi) + np.sum(np.log(covars), 1)
                  + np.sum((means ** 2) / covars, 1)
                  - 2 * np.dot(obs, (means / covars).T)
                  + np.dot(obs ** 2, (1.0 / covars).T))
    return lpr


def _lmvnpdfspherical(obs, means=0.0, covars=1.0):
    cv = covars.copy()
    if covars.ndim == 1:
        cv = cv[:, np.newaxis]
    return _lmvnpdfdiag(obs, means, np.tile(cv, (1, obs.shape[-1])))


def _lmvnpdftied(obs, means, covars):
    from scipy import linalg
    n_obs, n_dim = obs.shape
    # (x-y).T A (x-y) = x.T A x - 2x.T A y + y.T A y
    icv = linalg.pinv(covars)
    lpr = -0.5 * (n_dim * np.log(2 * np.pi) + np.log(linalg.det(covars))
                  + np.sum(obs * np.dot(obs, icv), 1)[:, np.newaxis]
                  - 2 * np.dot(np.dot(obs, icv), means.T)
                  + np.sum(means * np.dot(means, icv), 1))
    return lpr


def _lmvnpdffull(obs, means, covars):
    """
    Log probability for full covariance matrices.
    """
    from scipy import linalg
    import itertools
    if hasattr(linalg, 'solve_triangular'):
        # only in scipy since 0.9
        solve_triangular = linalg.solve_triangular
    else:
        # slower, but works
        solve_triangular = linalg.solve
    n_obs, n_dim = obs.shape
    nmix = len(means)
    log_prob = np.empty((n_obs, nmix))
    for c, (mu, cv) in enumerate(itertools.izip(means, covars)):
        cv_chol = linalg.cholesky(cv, lower=True)
        cv_log_det = 2 * np.sum(np.log(np.diagonal(cv_chol)))
        cv_sol = solve_triangular(cv_chol, (obs - mu).T, lower=True).T
        log_prob[:, c] = -.5 * (np.sum(cv_sol ** 2, axis=1) + \
                           n_dim * np.log(2 * np.pi) + cv_log_det)

    return log_prob


def _validate_covars(covars, cvtype, nmix, n_dim):
    from scipy import linalg
    if cvtype == 'spherical':
        if len(covars) != nmix:
            raise ValueError("'spherical' covars must have length nmix")
        elif np.any(covars <= 0):
            raise ValueError("'spherical' covars must be non-negative")
    elif cvtype == 'tied':
        if covars.shape != (n_dim, n_dim):
            raise ValueError("'tied' covars must have shape (n_dim, n_dim)")
        elif (not np.allclose(covars, covars.T)
              or np.any(linalg.eigvalsh(covars) <= 0)):
            raise ValueError("'tied' covars must be symmetric, "
                             "positive-definite")
    elif cvtype == 'diag':
        if covars.shape != (nmix, n_dim):
            raise ValueError("'diag' covars must have shape (nmix, n_dim)")
        elif np.any(covars <= 0):
            raise ValueError("'diag' covars must be non-negative")
    elif cvtype == 'full':
        if covars.shape != (nmix, n_dim, n_dim):
            raise ValueError("'full' covars must have shape "
                             "(nmix, n_dim, n_dim)")
        for n, cv in enumerate(covars):
            if (not np.allclose(cv, cv.T)
                or np.any(linalg.eigvalsh(cv) <= 0)):
                raise ValueError("component %d of 'full' covars must be "
                                 "symmetric, positive-definite" % n)


def _distribute_covar_matrix_to_match_cvtype(tiedcv, cvtype, n_states):
    if cvtype == 'spherical':
        cv = np.tile(np.diag(tiedcv).mean(), n_states)
    elif cvtype == 'tied':
        cv = tiedcv
    elif cvtype == 'diag':
        cv = np.tile(np.diag(tiedcv), (n_states, 1))
    elif cvtype == 'full':
        cv = np.tile(tiedcv, (n_states, 1, 1))
    else:
        raise (ValueError,
               "cvtype must be one of 'spherical', 'tied', 'diag', 'full'")
    return cv


def _covar_mstep_diag(gmm, obs, posteriors, avg_obs, norm, min_covar):
    # For column vectors:
    # covars_c = average((observation(t) - means_c) (observation(t) - means_c).T,
    #                    weights_c)
    # (observation(t) - means_c) (observation(t) - means_c).T
    #     = observation(t) observation(t).T - 2 observation(t) means_c.T + means_c means_c.T
    #
    # But everything here is a row vector, so all of the
    # above needs to be transposed.
    avg_obs2 = np.dot(posteriors.T, obs * obs) * norm
    avg_means2 = gmm._means ** 2
    avg_obs_means = gmm._means * avg_obs * norm
    return avg_obs2 - 2 * avg_obs_means + avg_means2 + min_covar


def _covar_mstep_spherical(*args):
    return _covar_mstep_diag(*args).mean(axis=1)


def _covar_mstep_full(gmm, obs, posteriors, avg_obs, norm, min_covar):
    # Eq. 12 from K. Murphy, "Fitting a Conditional Linear Gaussian
    # Distribution"
    cv = np.empty((gmm._n_states, gmm.n_features, gmm.n_features))
    for c in range(gmm._n_states):
        post = posteriors[:, c]
        avg_cv = np.dot(post * obs.T, obs) / post.sum()
        mu = gmm._means[c][np.newaxis]
        cv[c] = (avg_cv - np.dot(mu.T, mu)
                 + min_covar * np.eye(gmm.n_features))
    return cv


def _covar_mstep_tied2(*args):
    return _covar_mstep_full(*args).mean(axis=0)


def _covar_mstep_tied(gmm, obs, posteriors, avg_obs, norm, min_covar):
    #print "THIS IS BROKEN"
    # Eq. 15 from K. Murphy, "Fitting a Conditional Linear Gaussian
    avg_obs2 = np.dot(obs.T, obs)
    avg_means2 = np.dot(gmm._means.T, gmm._means)
    return (avg_obs2 - avg_means2 + min_covar * np.eye(gmm.n_features))


def _covar_mstep_slow(gmm, obs, posteriors, avg_obs, norm, min_covar):
    w = posteriors.sum(axis=0)
    covars = np.zeros(gmm._covars.shape)
    for c in range(gmm._n_states):
        mu = gmm._means[c]
        #cv = numpy.dot(mu.T, mu)
        avg_obs2 = np.zeros((gmm.n_features, gmm.n_features))
        for t, o in enumerate(obs):
            avg_obs2 += posteriors[t, c] * np.outer(o, o)
        cv = (avg_obs2 / w[c]
              - 2 * np.outer(avg_obs[c] / w[c], mu)
              + np.outer(mu, mu)
              + min_covar * np.eye(gmm.n_features))
        if gmm.cvtype == 'spherical':
            covars[c] = np.diag(cv).mean()
        elif gmm.cvtype == 'diag':
            covars[c] = np.diag(cv)
        elif gmm.cvtype == 'full':
            covars[c] = cv
        elif gmm.cvtype == 'tied':
            covars += cv / gmm._n_states
    return covars


_covar_mstep_funcs = {'spherical': _covar_mstep_spherical,
                      'diag': _covar_mstep_diag,
                      #'tied': _covar_mstep_tied,
                      'full': _covar_mstep_full,
                      'tied': _covar_mstep_slow,
                      }



# -*- coding: utf-8 -*-
#
# This file is part of s4d.
#
# s4d is a python package for speaker diarization.
# Home page: http://www-lium.univ-lemans.fr/s4d/
#
# s4d is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# s4d is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with s4d.  If not, see <http://www.gnu.org/licenses/>.


"""
Copyright 2014-2020 Sylvain Meignier
"""

import copy
import logging
import numpy

from .diar import Diar
from .utils import FeatureServerFake
from sidekit import Mixture, StatServer, FeaturesServer


class Viterbi:
    """
    Class that implements a Videtbi decoding and one-state HMM
    """
    eps = 1.0e-200

    def __init__(self, cep, diarization, exit_penalties=[0],
                 loop_penalties=[0]):
        self.cep = cep
        self.diarization = diarization
        shows = self.diarization.unique('show')
        self.cluster_list = self.diarization.unique('cluster')
        self.nb_clusters = len(self.cluster_list)
        if len(shows) > 1:
            raise Exception('diarization indexes serveal shows')

        self.nb_features = self.cep.shape[0]
        self.show = shows.pop()
        self.mixtures = None
        self.names = None
        self.observation = None
        self.transition_probabilities = None
        self.exit_penalties = exit_penalties
        self.loop_penalties = loop_penalties

    # def train_soft(self):
    #     self.mixtures = list()
    #     cluster_eatures = self.diarization.features_by_cluster(index_max=self.nb_features)
    #     # init GMM
    #     for i in range(0, self.nb_clusters):
    #         cluster = self.cluster_list[i];
    #         m = Mixture(name=cluster)
    #         idx = cluster_eatures[cluster]
    #         data = self.cep[idx]
    #         llk = m.EM_uniform(data, 8, 3, 10, llk_gain=0.01)
    #         self.mixtures.append(m)
    #     self._init_transition()
    #
    #     for it in range(1):
    #         # print('emission: ', len(self.mixtures), self.nb_clusters, len(self.cluster_list))
    #         self.emission()
    #         # soft train
    #         # print('soft train: ', len(self.mixtures), self.nb_clusters, len(self.cluster_list))
    #         for i in range(len(self.mixtures)):
    #             # print(it, i, self.cluster_list[i], self.nb_clusters)
    #             m = self.mixtures[i]
    #             # expectation
    #             accum = copy.deepcopy(m)
    #             accum._reset()
    #
    #             for seg in self.diarization:
    #                 #if seg['cluster'] != m.speaker:
    #                 #    continue
    #                 #else:
    #                 #    print(seg['cluster'], m.speaker)
    #                 start = seg['start']
    #                 stop = seg['stop']
    #                 llks = numpy.mean(self.observation[start:stop, :], axis=0)
    #                 w = llks / numpy.sum(llks)
    #                 # w = llks / numpy.max(llks)
    #                 # w = llks / 6.0
    #                 data = self.cep[start:stop]
    #
    #                 lp = m.compute_log_posterior_probabilities(data)
    #                 pp, loglk = sum_log_probabilities(lp)
    #                 pp *= w[i]
    #                 # zero order statistics
    #                 accum.w += pp.sum(0)
    #                 # print('\t', start, stop, accum.w)
    #                 # first order statistics
    #                 accum.mu += numpy.dot(data.T, pp).T
    #                 # second order statistics
    #                 accum.invcov += numpy.dot(numpy.square(data.T), pp).T
    #
    #             m.w = accum.w / numpy.sum(accum.w)
    #             m.mu = accum.mu / accum.w[:, numpy.newaxis]
    #             cov = accum.invcov / accum.w[:, numpy.newaxis] - numpy.square(m.mu)
    #             m.invcov = 1.0 / cov
    #             m._compute_all()

    def train(self, distrib_nb=8, init=None, max_it=4):
        """
        Trains one GMM for each cluster using EM.

        """

        iterations=[1, 2, 4, 4, 8]
        idx = int(numpy.log2(distrib_nb)) - 1
        iterations[idx] = max_it

        self.mixtures = list()
        cluster_features = self.diarization.features_by_cluster(
            maximum_length=self.nb_features)
        new_cluster_list = list()
        self.names = list()
        for i in range(0, self.nb_clusters):
            cluster = self.cluster_list[i]
            index = cluster_features[cluster]
            data = self.cep[index]

            if init is None:
                mixture = Mixture(name=cluster)
                llk = mixture.EM_split(FeatureServerFake(data), [self.show], distrib_nb=distrib_nb, iterations=iterations, llk_gain=0.01, num_thread=1)
            else:
                mixture = init[i]
                if mixture.name != cluster:
                    logging.error("!!! name don't match %s != %s", mixture.name, cluster)
                llk = mixture.EM_no_init(FeatureServerFake(data), [self.show], max_iteration=5,  llk_gain=0.01, num_thread=1)


            #llk = m.EM_uniform(FeatureServerFake(data), [self.show], distrib_nb=distrib_nb, llk_gain=0.01, num_thread=1)
            sum_llk = sum(llk)
            if numpy.isfinite(sum_llk) and sum_llk != 0.0:
                self.mixtures.append(mixture)
                self.names.append(self.mixtures[i].name)
                new_cluster_list.append(cluster)
            else:
                logging.warning('bad model, remove it: ' + cluster + ' '+ str(llk)+ ' nb features: '+str(len(index)))
        self.cluster_list = new_cluster_list
        self.nb_clusters = len(self.cluster_list)
        self._init_transition()

    def _init_transition(self):
        self.transition_probabilities = numpy.full((self.nb_clusters, self.nb_clusters),
                                                   self.exit_penalties[-1], dtype=numpy.int)
        for i in range(0, self.nb_clusters):
            self.transition_probabilities[i, i] = self.loop_penalties[
                min(i, len(self.loop_penalties) - 1)]
            if i < len(self.exit_penalties) - 1:
                for j in range(0, self.nb_clusters):
                    if i != j:
                        self.transition_probabilities[i, j] = self.exit_penalties[
                            min(i, len(self.exit_penalties) - 1)]

    def emission(self):
        """
        Computes the log-likelihood for each features.

        """
        self.observation = numpy.zeros((self.nb_features, self.nb_clusters))
        #corrupt_llk_list = list()
        for i in range(0, self.nb_clusters):
            lp = self.mixtures[i].compute_log_posterior_probabilities(self.cep)
            #self.observation[:, i] = numpy.log(numpy.sum(numpy.exp(lp), axis=1))
            pp_max = numpy.max(lp, axis=1)
            self.observation[:, i]  = pp_max + numpy.log(numpy.sum(numpy.exp((lp.transpose() - pp_max).transpose()), axis=1))
            #logging.info("--> %f %f", numpy.mean(self.observation[:, i]), numpy.mean(ll))


        #     finite = numpy.logical_not(numpy.isfinite(self.observation[:, i]))
        #     cpt = numpy.count_nonzero(finite)
        #
        #     if cpt >= self.nb_features/10:
        #         logging.debug('model ' + self.cluster_list[i] + '(' + str(i) + '), nb trame with llk problem: ' + str(cpt) + ' ' + str(self.nb_features))
        #         corrupt_llk_list.append(i)
        #     else:
        #         self.observation[finite, i] = numpy.finfo('d').min
        #
        # if len(corrupt_llk_list) > 0:
        #     for i in reversed(corrupt_llk_list):
        #         del self.cluster_list[i]
        #         del self.mixtures[i]
        #     self.nb_clusters = len(self.cluster_list)
        #
        #     self.observation = numpy.delete(self.observation, corrupt_llk_list, axis=1)
        #     self.transition_probabilities = numpy.delete(self.transition_probabilities, corrupt_llk_list, axis=1)
        #     self.transition_probabilities = numpy.delete(self.transition_probabilities, corrupt_llk_list, axis=0)

    def decode(self, table):
        """
        performs a Viterbi decoding of the segment given in diarization
        :param table: a Diar object
        :return: a Diar object
        """

        # print(self.transition_probabilities)
        # print(self.observation)

        path = numpy.ones((self.nb_features, self.nb_clusters), 'int32') * -1
        path[0, :] = numpy.arange(self.nb_clusters)
        out_diarization = Diar()

        for row in table:
            start = row['start']
            stop = min(row['stop'], self.nb_features-1)
            logging.debug('perform from %d to %d', start, stop)

            for t in range(start, stop+1):
                tmp = self.observation[t - 1, :] + self.transition_probabilities
                self.observation[t, :] += numpy.max(tmp, axis=1)
                path[t, :] = numpy.argmax(tmp, axis=1)

            max_pos = numpy.argmax(self.observation[stop, :])
            out_diarization.append(show=self.show, start=stop - 1, stop=stop,
                             cluster=self.cluster_list[max_pos])
            for t in range(stop - 1, start, -1):
                max_pos = path[t, max_pos]
                cluster = self.cluster_list[max_pos]
                if (out_diarization[-1]['start'] == t) and (
                    out_diarization[-1]['cluster'] == cluster):
                    out_diarization[-1]['start'] -= 1
                else:
                    out_diarization.append(show=self.show, start=t - 1, stop=t,
                                     cluster=cluster)
        out_diarization.sort()
        # self.observation = None
        return out_diarization


def viterbi_decoding(cep, diarization, penalty):
    init_diarization = copy.deepcopy(diarization)
    if len(init_diarization) <=1:
        return init_diarization
    for seg in init_diarization:
        seg['cluster'] = 'init'
    init_diarization.pack()
    hmm = Viterbi(cep, diarization, exit_penalties=[penalty])
    hmm.train()
    hmm.emission()
    return hmm.decode(init_diarization)


class ViterbiMap(Viterbi):
    eps = 1.0e-200

    def __init__(self, featureServer, diarization, ubm, exit_penalties=[0],
                 loop_penalties=[0], alpha=0.9, linear=False):
        assert isinstance(featureServer, FeaturesServer), 'First parameter should be a FeatureServer'

        self.featureServer = featureServer
        self.ubm = ubm
        self.alpha = alpha
        self.linear = linear
        self.diarization = diarization
        shows = self.diarization.unique('show')
        self.cluster_list = self.diarization.unique('cluster')
        self.nb_clusters = len(self.cluster_list)
        if len(shows) > 1:
            raise Exception('diarization indexes serveal shows')
        self.cep, lbl = self.featureServer.load(shows[0])

        self.nb_features = self.cep.shape[0]
        self.show = shows.pop()
        self.mixtures = None
        self.names = None
        self.observation = None
        self.observation_ubm = None
        self.transition_probabilities = None
        self.exit_penalties = exit_penalties
        self.loop_penalties = loop_penalties

    def train(self):
            idmap = self.diarization.id_map()
            stat=StatServer(idmap, self.ubm)
            stat.accumulate_stat(ubm=self.ubm, feature_server=self.featureServer, seg_indices=range(stat.segset.shape[0]), num_thread=1)
            stat = stat.sum_stat_per_model()[0]
            self.mixtures = stat.adapt_mean_MAP(self.ubm, self.alpha, linear=self.linear)
            self.names = self.mixtures.modelset
            #print(self.names)
            #print(self.mixtures.stat1[:, 0:24])
            self._init_transition()

    def emission(self, ubm=False):
        self.observation = numpy.zeros((self.nb_features, self.nb_clusters))
        self.observation_ubm = None
        if ubm:
            self.observation_ubm = numpy.zeros((self.nb_features, 1))
            lp = self.ubm.compute_log_posterior_probabilities(self.cep)
            #self.observation_ubm = numpy.log(numpy.sum(numpy.exp(lp), axis=1))
            pp_max = numpy.max(lp, axis=1)
            self.observation_ubm = pp_max + numpy.log(numpy.sum(numpy.exp((lp.transpose() - pp_max).transpose()), axis=1))

        for i in range(0, self.nb_clusters):
            logging.info('emission name: %s', self.mixtures.modelset[i])
            mean = self.mixtures.stat1[i, :]
            lp = self.ubm.compute_log_posterior_probabilities(self.cep, mean)
            #self.observation[:, i] = numpy.log(numpy.sum(numpy.exp(lp), axis=1))
            pp_max = numpy.max(lp, axis=1)
            self.observation[:, i] = pp_max + numpy.log(numpy.sum(numpy.exp((lp.transpose() - pp_max).transpose()), axis=1))
            #print(self.observation[0:10, i])



# -*- coding: utf-8 -*-
#
# This file is part of s4d.
#
# s4d is a python package for speaker diarization.
# Home page: http://www-lium.univ-lemans.fr/s4d/
#
# s4d is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# s4d is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with s4d.  If not, see <http://www.gnu.org/licenses/>.


"""
Copyright 2014-2020 Sylvain Meignier
"""

import copy
import numpy as np

from scipy.optimize import linear_sum_assignment


class DER:
    """ Class to compute a s4d error rate.
    Error is computed at a rate of 10ms.

    :attr uem_set: a set of index to evaluate
    :attr collar: is the no-score zone around reference speaker segment
    boundaries.  (Speaker Diarization output is not evaluated within +/- collar
    seconds of a reference speaker segment boundary.)
    :attr uem_set_collar: a set of index on with the collar is applied
    :attr length: the last index in the reference
    :attr ref_cluster_list: the list of reference cluster_list
    :attr ref_mat: the numpy.ndarry of reference indexes, each row correspond
    to a reference cluster (same order as ref_cluster_list)
    :attr hyp_cluster_list: the list of hypothesis cluster_list
    :attr hyp_mat: the numpy.ndarry of reference indexes, each row correspond
    to a hypothesis cluster (same order as ref_cluster_list)
    :attr conf: the confusion matrix
    :attr assigned: a numpy.ndarray given the association between the references and hypotheses
    :attr not_assigned: a list of hypothesis cluster_list not associated to a reference cluster

    """
    def __init__(self, hyp_diarization, ref_diarization, uem_diarization = None, collar=0, no_overlap=False):
        """

        :param hyp_diarization: a hypothesis Diar object
        :param ref_diarization: a reference Diar object
        :param uem_diarization: a uem Diar object
        :param collar: is the no-score zone around reference speaker segments
        :param no_overlap: to limit scoring to those time regions in which only
        a single speaker is speaking

        """
        uem = uem_diarization
        self.collar = collar
        self.match = None
        if uem_diarization is not None:
            uem = copy.deepcopy(uem_diarization)
            if len(uem.unique('cluster')) > 1:
                raise Exception('UEM contains more than one cluster')
            if len(uem.unique('show')) > 1:
                raise Exception('UEM contains more than one show')
            self.uem_set = set(uem.features())

        ref = copy.deepcopy(ref_diarization)
        ref.pack()
        lst = ref.unique('show')
        if len(lst) > 1:
            raise Exception('REF contains more than one show')
        else:
            self.show = lst[0]
        if uem_diarization is None:
            print('uem from ref')
            self.uem_set = set([i for i in range(min(ref.unique('start')), max(ref.unique('stop')))])

        if collar > 0:
            rem = list()
            for row in ref:
                rem += [i for i in range(row['start']-collar, row['start']+collar)]
                rem += [i for i in range(row['stop']-collar, row['stop']+collar)]
            self.uem_set_collar = self.uem_set - set(rem)
        else:
            self.uem_set_collar = self.uem_set

        self.length = max(self.uem_set_collar)+1
        self.ref_cluster_list = ref.unique('cluster')
        ref_idx = ref.features_by_cluster()
        self.ref_mat = self._to_bool(ref_idx, self.ref_cluster_list, self.uem_set_collar)
        if no_overlap:
            print('remove overlap')
            sum = np.sum(self.ref_mat, axis=0)
            idx = np.squeeze(np.asarray(np.argwhere(sum == 1))).tolist()
            self.uem_set_collar = self.uem_set & set(idx)
            self.length = max(self.uem_set_collar)+1
            self.ref_mat = self._to_bool(ref_idx, self.ref_cluster_list, self.uem_set_collar)

        hyp = copy.deepcopy(hyp_diarization)
        hyp.pack()
        self.hyp_cluster_list = hyp.unique('cluster')
        hyp_idx = hyp.features_by_cluster()
        self.hyp_mat = self._to_bool(hyp_idx, self.hyp_cluster_list, self.uem_set_collar)

        self.conf = None
        self.assigned = None
        self.not_assigned = None

    def _to_bool(self, map, cluster_list, uem_set):
        """Transform the dict of segments indexed by cluster into a matrix
        of booleans.

        :param map: dict of segments indexed by speaker
        :param cluster_list: list of cluster_list
        :param uem_set: the segments is filtered according to the uem


        :return: a numpy.ndarray object of shape (#cluster, #features)
        """
        mat = np.zeros((len(cluster_list), self.length))

        for i, cluster in enumerate(cluster_list):
            mat[i, list(set(map[cluster]) & uem_set)] = True
        return mat

    def confusion(self):
        """Compute the confusion matrix between reference and hypothesis

        :return: a numpy.ndarray object of shape (#reference, #hypothesis)
        """
        #nh = len(self.hyp_cluster_list)
        #nr = len(self.ref_cluster_list)
        #self.conf = np.full((nr, nh), 0.0)
        #for i in range(nr):
        #    for j in range(nh):
        #        self.conf[i, j] = np.sum(np.logical_and(self.ref_mat[i], self.hyp_mat[j]))

        self.conf = np.dot(self.ref_mat, self.hyp_mat.transpose())

        return self.conf

    def confusion2str(self, all_value=False):
        """Generate a string version of the confusion matrix

        :param all_value: if false, only the confision score upper to 0 is
        generated

        :return: a str
        """
        mr = mh = 0
        for ch in self.ref_cluster_list+['reference']:
            if mr < len(ch):
                mr = len(ch)
        for ch in self.hyp_cluster_list+['hypothesis']:
            if mh < len(ch):
                mh = len(ch)

        f = '{:'+str(mr)+'} {:'+str(mh)+'} {:>9}\n'
        line = f.format('reference', 'hypothesis', 'conf')
        f = '{:'+str(mr)+'} {:'+str(mh)+'} {:>8.2f}s\n'
        nh = len(self.hyp_cluster_list)
        nr = len(self.ref_cluster_list)

        for i in range(nr):
            for j in range(nh):
                if self.conf[i, j] > 0 or all_value == True:
                    r = self.ref_cluster_list[i]
                    h = self.hyp_cluster_list[j]
                    s = self.conf[i, j]/100
                    line += f.format(r, h, s)
        return line

    def assignment(self):
        """Compute the reference to hypothese association using the hungarian
        algorithm

        :return: a dict object with key equal to (reference cluster, hypothesis cluster)
        and value equal to the confusion score
        :return: a list object containing the hypothesis cluster_list that are not assigned
        """
        self.assigned = np.stack(linear_sum_assignment(-1.0 * self.conf), 1)
        name_assigned = dict()
        self.match = 0;
        self.not_assigned = copy.deepcopy(self.hyp_cluster_list)

        for i in range(self.assigned.shape[0]):
            r = self.ref_cluster_list[self.assigned[i, 0]]
            h = self.hyp_cluster_list[self.assigned[i, 1]]
            s = self.conf[self.assigned[i, 0], self.assigned[i, 1]]
            if s > 0:
                self.not_assigned.remove(h)
                name_assigned[(r, h)] = s
                self.match += s
            else:
                self.assigned[i, 0] = -1
                self.assigned[i, 1] = -1

        return name_assigned, self.not_assigned

    def set_assignment(self, name_assigned):
        """Set assignment from an external source (from a collection of cluster_list)

        :param: name_assigned is a dict object with key equal to
        (reference cluster, hypothesis cluster) and value equal to the confusion
        score
        """
        i = 0;
        self.assigned = np.ones((len(self.ref_cluster_list), 2)) * -1

        for key in name_assigned:
            # print(key)
            if key[0] in self.ref_cluster_list and key[1] in self.hyp_cluster_list:
                r = self.ref_cluster_list.index(key[0])
                h = self.hyp_cluster_list.index(key[1])
                self.assigned[i, 0] = r
                self.assigned[i, 1] = h
                i += 1

    def assignment2str(self):
        """Generate a string version of the assignment matrix attribut

        :return: a str
        """
        if self.assigned is None:
            return "Assignment is not set"
        mr = mh = 0

        for ch in self.ref_cluster_list+['reference']:
            if mr < len(ch):
                mr = len(ch)
        for ch in self.hyp_cluster_list+['hypothesis']:
            if mh < len(ch):
                mh = len(ch)
        f = '{:'+str(mr)+'} {:'+str(mh)+'} {:>9}\n'
        line = f.format('reference', 'hypothesis', 'matching')
        f = '{:'+str(mr)+'} {:'+str(mh)+'} {:>8.2f}s {:d} {:d}\n'

        for i in range(self.assigned.shape[0]):
            if self.assigned[i, 0] >= 0:
                r = self.ref_cluster_list[self.assigned[i, 0]]
                h = self.hyp_cluster_list[self.assigned[i, 1]]
                s = self.conf[self.assigned[i, 0], self.assigned[i, 1]]/100
                line += f.format(r, h, s, self.assigned[i, 0], self.assigned[i, 1])

        r = 'None'
        s = 0.0
        for h in self.not_assigned:
            line += f.format(r, h, s, -1, -1)

        return line

    def _correct(self):
        """compute the correct decision

        :return: a numpy.ndarray
        """
        c = np.zeros(self.length)
        for i in range(self.assigned.shape[0]):
            if self.assigned[i, 0] >= 0 and self.assigned[i, 1] >= 0:
                diff = np.logical_and(self.ref_mat[self.assigned[i, 0]], self.hyp_mat[self.assigned[i, 1]])
                c += diff
        return c

    def error(self):
        """compute the s4d error

        :return: a DER_result object
        """
        h_sum = np.sum(self.hyp_mat, axis=0)
        r_sum = np.sum(self.ref_mat, axis=0)
        c = self._correct()

        res = DER_result(self.show)
        res.uem_time = len(self.uem_set)
        res.uem_with_collar_time = len(self.uem_set_collar)
        res.uem_collar_time = res.uem_time - res.uem_with_collar_time

        z = np.zeros(r_sum.shape[0])

        mrh = np.sum(np.maximum(r_sum - h_sum, z))
        mhr = np.sum(np.maximum(h_sum - r_sum, z))

        res.speech_time = np.count_nonzero(r_sum)
        res.sns_miss_time = np.count_nonzero(np.logical_and(r_sum, np.logical_not(h_sum)))
        res.sns_fa_time = np.count_nonzero(np.logical_and(h_sum, np.logical_not(r_sum)))
        res.spk_time = r_sum.sum()
        res.spk_miss_time = mrh
        res.spk_fa_time = mhr
        res.spk_conf_time = np.sum(np.minimum(h_sum, r_sum) - c)

        # for i in range(c.shape[0]):
        #     ref = r_sum[i]
        #     hyp = h_sum[i]
        #     mrh = max(ref - hyp, 0)
        #     mhr = max(hyp - ref, 0)
        #     res.speech_time += bool(ref)
        #     res.sns_miss_time += bool(ref and not hyp)
        #     res.sns_fa_time += bool(hyp and not ref)
        #     if ref > 1:
        #         res.overlap_time += ref
        #         res.overlap_miss_time += mrh
        #         res.overlap_fa_time += mhr
        #     res.spk_time += ref
        #     res.spk_miss_time += mrh
        #     res.spk_fa_time += mhr
        #     res.spk_conf_time += min(hyp, ref) - c[i]
        res.compute_rate()
        return res

class DER_coll(DER):
    def __init__(self, ref_cluster_list, hyp_cluster_list):
        self.ref_cluster_list = ref_cluster_list
        self.hyp_cluster_list = hyp_cluster_list
        hyp_nb_clusters = len(hyp_cluster_list)
        ref_nb_clusters = len(ref_cluster_list)
        self.conf = np.full((ref_nb_clusters, hyp_nb_clusters), 0.0)

    def merge_confusion(self, der):
        for i in range(len(der.ref_cluster_list)):
            for j in range(len(der.hyp_cluster_list)):
                v = der.conf[i, j]
                if v > 0:
                    ig = self.ref_cluster_list.index(der.ref_cluster_list[i])
                    jg = self.hyp_cluster_list.index(der.hyp_cluster_list[j])
                    self.conf[ig, jg] += v

class DER_result:
    """
    Class to store the s4d error results
    """
    def __init__(self, show):
        """

        :param show: a string
        """
        self.show = show

        self.uem_time = 0
        self.uem_with_collar_time = 0
        self.uem_collar_time = 0
        self.spk_time = 0
        self.speech_time = 0
        self.sns_time = 0
        self.overlap_time = 0
        self.overlap_miss_time = 0
        self.overlap_fa_time = 0
        self.spk_conf_time = 0
        self.sns_fa_time = 0
        self.sns_miss_time = 0
        self.overlap_miss = 0
        self.overlap_fa = 0
        self.spk_miss_time = 0
        self.spk_fa_time = 0
        self.spk_miss = 0
        self.spk_fa = 0
        self.spk_conf = 0
        self.speech_fa = 0
        self.speech_miss = 0
        self.len_line = 0

    def accumulate(self, result):
        """
        Accumulate the scores
        :param result: a DER_result object to accumulate
        :return:
        """
        self.uem_time += result.uem_time
        self.uem_with_collar_time += result.uem_with_collar_time
        self.uem_collar_time += result.uem_collar_time
        self.spk_time += result.spk_time
        self.speech_time += result.speech_time
        self.sns_time += result.sns_time
        self.overlap_time += result.overlap_time
        self.overlap_miss_time += result.overlap_miss_time
        self.overlap_fa_time += result.overlap_fa_time
        self.spk_conf_time += result.spk_conf_time
        self.sns_fa_time += result.sns_fa_time
        self.sns_miss_time += result.sns_miss_time
        self.overlap_miss += result.overlap_miss
        self.overlap_fa += result.overlap_fa
        self.spk_miss_time += result.spk_miss_time
        self.spk_fa_time += result.spk_fa_time
        self.spk_miss += result.spk_miss
        self.spk_fa += result.spk_fa
        self.spk_conf += result.spk_conf
        self.speech_fa += result.speech_fa
        self.speech_miss += result.speech_miss

    def compute_rate(self):
        """
        Compute the various rates
        """
        self.sns_fa = self.sns_fa_time / self.uem_time * 100
        self.sns_miss = self.sns_miss_time / self.uem_time * 100

        if self.overlap_time > 0:
            self.overlap_fa = self.overlap_fa_time / self.overlap_time * 100
            self.overlap_miss = self.overlap_miss_time / self.overlap_time * 100
        else:
            self.overlap_fa = 0
            self.overlap_miss = 0

        self.spk_fa = self.spk_fa_time / self.spk_time * 100
        self.spk_miss = self.spk_miss_time / self.spk_time * 100
        self.spk_conf = self.spk_conf_time / self.spk_time * 100

    @classmethod
    def header(cls, speech=True, overlap=False, speaker=True):
        """
        Make the header
        :param speech: add speech/non-speech results
        :param overlap: add overlap result
        :param speaker: add speaker result (DER)
        :return: a str
        """
        line = ['show', 'type']
        if speech:
            line += ['fa', 'miss', 'sns']
        if overlap:
            line += ['fa', 'miss', 'overlap']
        if speaker:
            line += ['fa', 'miss', 'conf', 'speaker']

        return [line]

    def rate(self, name=None, speech=True, overlap=False, speaker=True):
        """
        Make the rate results
        :param speech: add speech/non-speech results
        :param overlap: add overlap result
        :param speaker: add speaker result (DER)
        :return: a str
        """
        if name is None:
            name = ''
        line = [name, 'rate']
        if speech:
            line += [self.sns_fa, self.sns_miss, self.sns_fa+self.sns_miss]
        if overlap:
            line += [self.overlap_fa, self.overlap_miss, self.overlap_fa+self.overlap_miss]
        if speaker:
            line += [self.spk_fa, self.spk_miss, self.spk_conf, self.get_der()]
        return [line]

    def time(self, name=None, speech=True, overlap=False, speaker=True):
        """
        Make the time results
        :param speech: add speech/non-speech results
        :param overlap: add overlap result
        :param speaker: add speaker result (DER)
        :return: a str
        """
        if name is None:
            name = ''
        line = [name, 'time']
        if speech:
            line += [self.sns_fa_time/100, self.sns_miss_time/100, self.uem_with_collar_time/100]
        if overlap:
            line += [self.overlap_fa_time/100, self.overlap_miss_time/100, self.overlap_time/100]
        if speaker:
            line += [self.spk_fa_time/100, self.spk_miss_time/100, self.spk_conf_time/100, self.spk_time/100]
        return [line]

    def get_table(self, name=None, time=True, rate=True, overlap=False):
        lst = []
        add_name = False
        if rate:
            lst += self.rate(name=name, overlap=overlap)
            add_name = True

        if time:
            if not add_name:
                lst += self.time(name=name, overlap=overlap)
            else:
                lst += self.time(name=None)

        #if add_name == False and overlap:
        #    lst += self.overlap(speaker=speaker)
        #    add_name = True
        #else:
            #lst += self.overlap(speaker=None)
        return lst

    def get_der(self):
        return self.spk_fa+self.spk_miss+self.spk_conf


def get_header():
    return DER_result.header()


def compute_der(hyp, ref, uem=None, collar=0, no_overlap=True):
    cder = DER(hyp, ref, uem, collar=collar, no_overlap=no_overlap)
    cder.confusion()
    cder.assignment()
    return cder.error()



# -*- coding: utf-8 -*-
#
# This file is part of s4d.
#
# s4d is a python package for speaker diarization.
# Home page: http://www-lium.univ-lemans.fr/s4d/
#
# s4d is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# s4d is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with s4d.  If not, see <http://www.gnu.org/licenses/>.


"""
Copyright 2014-2020 Sylvain Meignier
"""

import copy
import logging
import numpy as np

from scipy.spatial.distance import squareform
from scipy.cluster import hierarchy as hac
from .hac_utils import *


def information(merge, nb_merge, i, j, value):
    merge.append([nb_merge, i, j, value])


def hac_iv(diar, scores, method="complete", threshold=0.0):
    ldiar = copy.deepcopy(diar)
    lscores = copy.deepcopy(scores)
    # get the triangular part of the distances
    distances, t = scores2distance(lscores, threshold)

    # distance = numpy.copy((scores.scoremat + scores.scoremat.T) / 2.0) * -1.0
    # numpy.fill_diagonal(distance, numpy.inf)
    # min = numpy.min(distance)
    # distance -= min
    # numpy.fill_diagonal(distance, 0.0)
    distance_sym = squareform(distances)
    # t = -1.0 * threshold - min
    # cluster the data
    link = hac.linkage(distance_sym, method=method)
    # print(link)
    # assign new cluster
    # d : 'key' give the new names of cluster_list in values (a list)
    cluster_dict = dict()
    merge = list()
    i = 0
    cluster_list = scores.modelset.tolist()
    # print(cluster_list)
    while i < len(link) and link[i, 2] < t:
        # the cluster_list of the 2 clusters
        logging.debug('c0: {:d} c1: {:d} value: {:.4f}'.format(int(link[i, 0]),
                                                              int(link[i, 1]),
                                                              link[i, 2]))
        c0 = cluster_list[int(link[i, 0])]
        c1 = cluster_list[int(link[i, 1])]
        logging.debug(
            '\t c0: {} c1: {} value: {:.4f}'.format(c0, c1, link[i, 2]))
        information(merge, i, c0, c1, link[i, 2])
        if c1 in cluster_dict:
            # c0 is put in c1, and c1 is not empty
            cluster_dict[c1].append(c0)
        else:
            cluster_dict[c1] = [c0]
        if c0 in cluster_dict:
            # remove c0 key
            cluster_dict[c1] += cluster_dict[c0]
            cluster_dict.pop(c0)
        # add the speaker of the new cluster
        cluster_list.append(c1)
        ldiar.rename('cluster', [c0], c1)
        i += 1

    return ldiar, cluster_dict, merge

# def hac_update_model(diarization, ivectors):


def hac_iv_it(diar, model_iv, threshold=0.0):
    model_iv_local = copy.deepcopy(model_iv)
    model_iv_local.diar = copy.deepcopy(diar)
    np.fill_diagonal(model_iv_local.scores.scoremat, -np.inf)

    nb = model_iv_local.scores.modelset.shape[0]

    i, j, v = argmax(model_iv_local.scores.scoremat, nb)
    nb_merge = 0
    while v > threshold and nb > 1:
        nb_merge += 1
        logging.info('merge: %d c1: %s (%d) c2: %s (%d) dist: %f, size: %d',
                     nb_merge, model_iv_local.scores.modelset[i], i,
                     model_iv_local.scores.modelset[j], j,
                     v, model_iv_local.scores.modelset.shape[0])
        name_i = model_iv_local.scores.modelset[i]
        name_j = model_iv_local.scores.modelset[j]

        model_iv_local.update(i, j)
        model_iv_local.diar.rename('cluster', [name_j], name_i)
        np.fill_diagonal(model_iv_local.scores.scoremat, -np.inf)

        nb = model_iv_local.scores.modelset.shape[0]
        i, j, v = argmax(model_iv_local.scores.scoremat, nb)


    return model_iv_local.diar




# -*- coding: utf-8 -*-
#
# This file is part of s4d.
#
# s4d is a python package for speaker diarization.
# Home page: http://www-lium.univ-lemans.fr/s4d/
#
# s4d is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# s4d is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with s4d.  If not, see <http://www.gnu.org/licenses/>.


"""
Copyright 2014-2020 Sylvain Meignier
"""

import copy
import numpy as np
import scipy.cluster.hierarchy as hac

from matplotlib import pyplot as plot
from matplotlib import ticker


def plot_dendrogram(merge, thr, size=(25,6), log=False):
    """

    :param merge:
    :param thr:
    :param size:
    :param log:
    :return:
    """
    minv = 0

    def my_formatter(x, pos):
        """

        :param x:
        :param pos:
        :return:
        """
        v = x
        if log:
            v = np.exp(v)
        v -= minv

        return "{:10.3f}".format(v)

    def link(merge, log):
        """

        :param merge:
        :param log:
        :return:
        """
        cluster_lst = list()
        idx = dict()
        qt = dict()
        k=0;
        while merge[k][0] < 0:
            name = merge[k][1]
            cluster_lst.append(name)
            idx[name] = k
            qt[name] = 1
            k += 1

        l = k
        links = np.zeros((l-1, 4))

        while k < len(merge):
            m = merge[k][0]
            name_i = merge[k][1]
            name_j = merge[k][2]
            v_ij = merge[k][3]

            qt[name_i] += 1
            links[m, 0] = idx[name_i]
            links[m, 1] = idx[name_j]
            links[m, 2] = v_ij
            links[m, 3] = qt[name_i]
            idx[name_i] = m+l
            idx[name_j] = -1

            k+= 1

        min = -np.min(links[:,2])+2
        links[:,2] += min

        if log:
            links[:,2] = np.log(links[:,2])

        return cluster_lst, links, min

    merge = copy.deepcopy(merge)

    plot.figure(figsize=size)
    cluster_list, link_mat, minv = link(merge, log)

    t=thr+minv
    if log:
        t = np.log(t)

    dendro_data = hac.dendrogram(link_mat, labels=cluster_list, color_threshold=t)
    for i, d in zip(dendro_data['icoord'], dendro_data['dcoord']):
        x = 0.5 * sum(i[1:3])
        y = d[1]
        v = y
        if log:
            v = np.exp(v)
        v -= minv
        plot.plot(x, y, 'o', c="b")
        plot.annotate("{:10.3f}".format(v), (x, y), xytext=(0, -5),
                      textcoords='offset points',
                      va='top', ha='center')

    plot.axhline(y=t, c='r')
    plot.annotate("{:10.3f}".format(thr), (0, t), xytext=(25, 25),
                  textcoords='offset points',
                  va='top', ha='center')

    ax = plot.gca()
    ax.yaxis.set_major_formatter(ticker.FuncFormatter(my_formatter))

    plot.plot()

    return link_mat, dendro_data